#ifndef UTIL_H
#define UTIL_H

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <json-c/json.h>

#include "client.h"

#define WM_VERT_LEFT  0
#define WM_VERT_RIGHT 1
#define WM_HORIZ_TOP  0
#define WM_HORIZ_BOT  1

typedef enum NodeType {
    NODE_VERTICAL,
    NODE_HORIZONTAL,
    NODE_TAB,
    NODE_CLIENT,
    NODE_COUNT
} NodeType;

typedef struct TreeNode TreeNode;

typedef struct {
    TreeNode **nodes;
    size_t size;
    size_t capacity;
} NodeArray;

typedef struct {
    void **ptrs;
    size_t size;
    size_t capacity;
} PtrArray;

typedef struct {
    int x;
    int y;
    int w;
    int h;
} Rect;

struct TreeNode {
    NodeType type;

    TreeNode *parent;

    NodeArray *children;

    Rect pos;

    Client *client;
    unsigned int id;
};

typedef struct {
    size_t ws_index;
    char *str;
} WmWorkspaceToStrRet;

NodeArray* wm_nodearray_new();
void wm_nodearray_push(NodeArray *arr, TreeNode *node);
bool wm_nodearray_pop(NodeArray *arr, TreeNode **ret);
bool wm_nodearray_pop_front(NodeArray *arr, TreeNode **ret);
void wm_nodearray_clear(NodeArray *arr);
bool wm_nodearray_remove(NodeArray *arr, size_t index);
void wm_nodearray_free(NodeArray *arr);

TreeNode* wm_treenode_new(NodeType type, TreeNode *parent);
void wm_treenode_free(TreeNode *node);
bool wm_treenode_is_empty(TreeNode *node);
void wm_treenode_split_space(TreeNode *node, Rect *ret1, Rect *ret2);
void wm_treenode_recalculate_space(TreeNode *node);
void wm_treenode_swap(TreeNode *root, TreeNode *node1, TreeNode* node2);
TreeNode* wm_treenode_remove_client(Wm *wm, TreeNode *root, Client *client);
void wm_treenode_remove_node(Wm *wm, TreeNode *root, TreeNode *node);
int wm_get_node_index(TreeNode *parent, unsigned int node_id);
void wm_tree_to_DOT(TreeNode *root, const char *filename);
void wm_node_type_to_str(NodeType type, char *buf, size_t bufsize);
char* wm_treenode_to_str(TreeNode *node);
void wm_treenode_print(TreeNode *node);

PtrArray wm_ptrarray_new();
void wm_ptrarray_push(PtrArray *arr, void* ptr);
bool wm_ptrarray_pop(PtrArray *arr, void **ret);
bool wm_ptrarray_pop_front(PtrArray *arr, void **ret);
bool wm_ptrarray_remove(PtrArray *arr, size_t index);
void wm_ptrarray_free(PtrArray *arr);
TreeNode* wm_treenode_ptr_find_focused_client_node(TreeNode *root);
TreeNode* wm_treenode_ptr_find_client_node(TreeNode *root, Client *client);
NodeArray* wm_treenode_find_client_nodes(TreeNode *root);
TreeNode* wm_treenode_split_get_sibling(TreeNode *node);
NodeArray* wm_all_nodes_to_array(TreeNode *root);
void wm_log_state(Wm *wm, const char *prefixstr, const char* logfile);

TreeNode* wm_treenode_lmd(TreeNode *node);
NodeArray* wm_treenode_all_lmds(TreeNode *node);
NodeArray* wm_postorder_traversal(TreeNode *tree);
bool wm_is_treenode_keyroot(TreeNode *node);
NodeArray* wm_treenode_all_keyroots(TreeNode *node);

#endif
